@@ -39,6 +39,8 @@ class CustomPlot:
         The ylim of the axes as specified in matplotlib
     bounds: Bounds
         The bounds to show on the graph
+    node_idx : list
+        The node time to be plotted on the graphs
     parameters: Any
         The parameters of the function
     """
@@ -54,6 +56,9 @@ class CustomPlot:
         linestyle: str = None,
         ylim: Union[tuple, list] = None,
         bounds: Bounds = None,
+        node_idx: list = None,
+        label: list = None,
+        compute_derivative: bool = False,
         **parameters: Any,
     ):
         """
@@ -75,8 +80,14 @@ class CustomPlot:
             The style of the line as specified in matplotlib
         ylim: Union[tuple[float, float], list[float, float]]
             The ylim of the axes as specified in matplotlib
-        bounds:
+        bounds: Bounds
             The bounds to show on the graph
+        node_idx: list
+            The node time to be plotted on the graphs
+        label: list
+            Label of the curve to plot (to be added to the legend)
+        compute_derivative: bool
+            If the function should send the next node with x and u. Prevents from computing all at once (therefore a bit slower)
         """
 
         self.function = update_function
@@ -95,6 +106,9 @@ class CustomPlot:
         self.linestyle = linestyle
         self.ylim = ylim
         self.bounds = bounds
+        self.node_idx = node_idx
+        self.label = label
+        self.compute_derivative = compute_derivative
         self.parameters = parameters
 
 
@@ -102,7 +116,7 @@ class PlotOcp:
     """
     Attributes
     ----------
-    adapt_graph_size_to_bounds: bool
+    show_bounds: bool
         If the plot should adapt to bounds or to ydata
     all_figures: list
         The list of handlers to the matplotlib figures
@@ -181,8 +195,9 @@ class PlotOcp:
         self,
         ocp,
         automatically_organize: bool = True,
-        adapt_graph_size_to_bounds: bool = False,
+        show_bounds: bool = False,
         shooting_type: Shooting = Shooting.MULTIPLE,
+        use_scipy_integrator: bool = False,
     ):
         """
         Prepares the figures during the simulation
@@ -193,10 +208,12 @@ class PlotOcp:
             A reference to the ocp to show
         automatically_organize: bool
             If the figures should be spread on the screen automatically
-        adapt_graph_size_to_bounds: bool
+        show_bounds: bool
             If the axes should fit the bounds (True) or the data (False)
         shooting_type: Shooting
             The type of integration method
+        use_scipy_integrator: bool
+            Use the scipy solve_ivp integrator for RungeKutta 45 instead of currently defined integrator
         """
         for i in range(1, ocp.n_phases):
             if len(ocp.nlp[0].states["q"]) != len(ocp.nlp[i].states["q"]):
@@ -207,6 +224,7 @@ class PlotOcp:
             "general_options": {"use_tight_layout": False},
             "non_integrated_plots": {"linestyle": "-", "markersize": 3, "linewidth": 1.1},
             "integrated_plots": {"linestyle": "-", "markersize": 3, "linewidth": 1.1},
+            "point_plots": {"linestyle": None, "marker": ".", "markersize": 15},
             "bounds": {"color": "k", "linewidth": 0.4, "linestyle": "-"},
             "grid": {"color": "k", "linestyle": "-", "linewidth": 0.15},
             "vertical_lines": {"color": "k", "linestyle": "--", "linewidth": 1.2},
@@ -217,6 +235,7 @@ class PlotOcp:
 
         self.t = []
         self.t_integrated = []
+        self.use_scipy_integrator = use_scipy_integrator
         if isinstance(self.ocp.original_phase_time, (int, float)):
             self.tf = [self.ocp.original_phase_time]
         else:
@@ -243,7 +262,7 @@ class PlotOcp:
 
         self.plot_func = {}
         self.variable_sizes = []
-        self.adapt_graph_size_to_bounds = adapt_graph_size_to_bounds
+        self.show_bounds = show_bounds
         self.__create_plots()
         self.shooting_type = shooting_type
 
@@ -275,12 +294,16 @@ class PlotOcp:
         self.t_integrated = []
         last_t = 0
         for phase_idx, nlp in enumerate(self.ocp.nlp):
-            n_int_steps = nlp.ode_solver.steps
+            n_int_steps = nlp.ode_solver.steps_scipy if self.use_scipy_integrator else nlp.ode_solver.steps
             dt_ns = self.tf[phase_idx] / nlp.ns
             time_phase_integrated = []
             last_t_int = copy(last_t)
             for _ in range(nlp.ns):
-                time_phase_integrated.append(np.linspace(last_t_int, last_t_int + dt_ns, n_int_steps + 1))
+                if nlp.ode_solver.is_direct_collocation and not self.use_scipy_integrator:
+                    time_phase_integrated.append(np.array(nlp.dynamics[0].step_time) * dt_ns + last_t_int)
+                else:
+                    time_phase_integrated.append(np.linspace(last_t_int, last_t_int + dt_ns, n_int_steps + 1))
+
                 last_t_int += dt_ns
             self.t_integrated.append(time_phase_integrated)
 
@@ -294,6 +317,12 @@ class PlotOcp:
         Setup the plots
         """
 
+        def legend_without_duplicate_labels(ax):
+            handles, labels = ax.get_legend_handles_labels()
+            unique = [(h, l) for i, (h, l) in enumerate(zip(handles, labels)) if l not in labels[:i]]
+            if unique:
+                ax.legend(*zip(*unique))
+
         variable_sizes = []
         for i, nlp in enumerate(self.ocp.nlp):
             variable_sizes.append({})
@@ -305,9 +334,10 @@ class PlotOcp:
                         size = (
                             nlp.plot[key]
                             .function(
-                                np.zeros((nlp.states.shape, 1)),
-                                np.zeros((nlp.controls.shape, 1)),
-                                np.zeros((len(nlp.parameters), 1)),
+                                np.nan,
+                                np.zeros((nlp.states.shape, 2)),
+                                np.zeros((nlp.controls.shape, 2)),
+                                np.zeros((len(nlp.parameters), 2)),
                                 **nlp.plot[key].parameters,
                             )
                             .shape[0]
@@ -343,7 +373,6 @@ class PlotOcp:
                     axes = self.__add_new_axis(variable, nb, n_rows, n_cols)
                     self.axes[variable] = [nlp.plot[variable], axes]
 
-                t = self.t[i]
                 if variable not in self.plot_func:
                     self.plot_func[variable] = [
                         nlp_tp.plot[variable] if variable in nlp_tp.plot else None for nlp_tp in self.ocp.nlp
@@ -360,59 +389,94 @@ class PlotOcp:
                     ax.set_xlim(0, self.t[-1][-1])
                     if nlp.plot[variable].ylim:
                         ax.set_ylim(nlp.plot[variable].ylim)
-                    elif self.adapt_graph_size_to_bounds and nlp.plot[variable].bounds:
+                    elif self.show_bounds and nlp.plot[variable].bounds:
                         if nlp.plot[variable].bounds.type != InterpolationType.CUSTOM:
-                            y_min = nlp.plot[variable].bounds.min[ctr].min()
-                            y_max = nlp.plot[variable].bounds.max[ctr].max()
+                            y_min = nlp.plot[variable].bounds.min[ctr, :].min()
+                            y_max = nlp.plot[variable].bounds.max[ctr, :].max()
                         else:
                             nlp.plot[variable].bounds.check_and_adjust_dimensions(len(mapping), nlp.ns)
                             y_min = min([nlp.plot[variable].bounds.min.evaluate_at(j)[k] for j in range(nlp.ns)])
                             y_max = max([nlp.plot[variable].bounds.max.evaluate_at(j)[k] for j in range(nlp.ns)])
                         y_range, _ = self.__compute_ylim(y_min, y_max, 1.25)
                         ax.set_ylim(y_range)
-                    zero = np.zeros((t.shape[0], 1))
                     plot_type = self.plot_func[variable][i].type
+
+                    t = self.t[i][nlp.plot[variable].node_idx] if plot_type == PlotType.POINT else self.t[i]
+                    if self.plot_func[variable][i].label:
+                        label = self.plot_func[variable][i].label
+                    else:
+                        label = None
+
                     if plot_type == PlotType.PLOT:
+                        zero = np.zeros((t.shape[0], 1))
                         color = self.plot_func[variable][i].color if self.plot_func[variable][i].color else "tab:green"
                         self.plots.append(
                             [
                                 plot_type,
                                 i,
-                                ax.plot(t, zero, color=color, zorder=0, **self.plot_options["non_integrated_plots"])[0],
+                                ax.plot(
+                                    t,
+                                    zero,
+                                    color=color,
+                                    zorder=0,
+                                    label=label,
+                                    **self.plot_options["non_integrated_plots"],
+                                )[0],
                             ]
                         )
                     elif plot_type == PlotType.INTEGRATED:
-                        color = self.plot_func[variable][i].color if self.plot_func[variable][i].color else "tab:brown"
                         plots_integrated = []
-                        n_int_steps = nlp.ode_solver.steps
+                        n_int_steps = nlp.ode_solver.steps_scipy if self.use_scipy_integrator else nlp.ode_solver.steps
+                        zero = np.zeros(n_int_steps + 1)
+                        color = self.plot_func[variable][i].color if self.plot_func[variable][i].color else "tab:brown"
                         for cmp in range(nlp.ns):
                             plots_integrated.append(
                                 ax.plot(
                                     self.t_integrated[i][cmp],
-                                    np.zeros(n_int_steps + 1),
+                                    zero,
                                     color=color,
+                                    label=label,
                                     **self.plot_options["integrated_plots"],
                                 )[0]
                             )
                         self.plots.append([plot_type, i, plots_integrated])
-
                     elif plot_type == PlotType.STEP:
+                        zero = np.zeros((t.shape[0], 1))
                         color = self.plot_func[variable][i].color if self.plot_func[variable][i].color else "tab:orange"
                         linestyle = (
                             self.plot_func[variable][i].linestyle if self.plot_func[variable][i].linestyle else "-"
                         )
                         self.plots.append(
-                            [plot_type, i, ax.step(t, zero, linestyle, where="post", color=color, zorder=0)[0]]
+                            [
+                                plot_type,
+                                i,
+                                ax.step(t, zero, linestyle, where="post", color=color, zorder=0, label=label)[0],
+                            ]
+                        )
+                    elif plot_type == PlotType.POINT:
+                        zero = np.zeros((t.shape[0], 1))
+                        color = self.plot_func[variable][i].color if self.plot_func[variable][i].color else "tab:purple"
+                        self.plots.append(
+                            [
+                                plot_type,
+                                i,
+                                ax.plot(
+                                    t, zero, color=color, zorder=0, label=label, **self.plot_options["point_plots"]
+                                )[0],
+                                variable,
+                            ]
                         )
                     else:
                         raise RuntimeError(f"{plot_type} is not implemented yet")
 
+                    legend_without_duplicate_labels(ax)
+
                 for j, ax in enumerate(axes):
                     intersections_time = self.find_phases_intersections()
                     for time in intersections_time:
                         self.plots_vertical_lines.append(ax.axvline(time, **self.plot_options["vertical_lines"]))
 
-                    if nlp.plot[variable].bounds and self.adapt_graph_size_to_bounds:
+                    if nlp.plot[variable].bounds and self.show_bounds:
                         if nlp.plot[variable].bounds.type == InterpolationType.EACH_FRAME:
                             ns = nlp.plot[variable].bounds.min.shape[1] - 1
                         else:
@@ -517,11 +581,15 @@ class PlotOcp:
         self.ydata = []
 
         sol = Solution(self.ocp, v)
-        continuous = sum([nlp.ode_solver.is_direct_collocation for nlp in self.ocp.nlp]) > 0
-        data_states = sol.integrate(continuous=continuous, shooting_type=self.shooting_type, keepdims=False).states
+        data_states = sol.integrate(
+            continuous=False,
+            shooting_type=self.shooting_type,
+            keep_intermediate_points=True,
+            use_scipy_integrator=self.use_scipy_integrator,
+        ).states
         data_controls = sol.controls
         data_params = sol.parameters
-        data_params_in_dyn = np.array([data_params[key] for key in data_params if key != "time"]).squeeze()
+        data_params_in_dyn = np.array([data_params[key] for key in data_params if key != "all"]).reshape(-1, 1)
 
         for _ in self.ocp.nlp:
             if self.t_idx_to_optimize:
@@ -530,7 +598,7 @@ class PlotOcp:
             self.__update_xdata()
 
         for i, nlp in enumerate(self.ocp.nlp):
-            step_size = nlp.ode_solver.steps + 1
+            step_size = nlp.ode_solver.steps_scipy + 1 if self.use_scipy_integrator else nlp.ode_solver.steps + 1
             n_elements = nlp.ns * step_size + 1
 
             state = np.ndarray((0, n_elements))
@@ -546,28 +614,39 @@ class PlotOcp:
                 else:
                     control = np.concatenate((control, data_controls[s]))
 
-            if nlp.control_type == ControlType.CONSTANT:
-                u_mod = 1
-            elif nlp.control_type == ControlType.LINEAR_CONTINUOUS:
-                u_mod = 2
-            else:
-                raise NotImplementedError(f"Plotting {nlp.control_type} is not implemented yet")
-
             for key in self.variable_sizes[i]:
                 if not self.plot_func[key][i]:
                     continue
+                x_mod = 1 if self.plot_func[key][i].compute_derivative else 0
+                u_mod = (
+                    1
+                    if nlp.control_type == ControlType.LINEAR_CONTINUOUS or self.plot_func[key][i].compute_derivative
+                    else 0
+                )
 
                 if self.plot_func[key][i].type == PlotType.INTEGRATED:
                     all_y = []
                     for idx, t in enumerate(self.t_integrated[i]):
                         y_tp = np.empty((self.variable_sizes[i][key], len(t)))
                         y_tp.fill(np.nan)
-                        y_tp[:, :] = self.plot_func[key][i].function(
-                            state[:, step_size * idx : step_size * (idx + 1)],
-                            control[:, idx : idx + u_mod],
+
+                        val = self.plot_func[key][i].function(
+                            idx,
+                            state[:, step_size * idx : step_size * (idx + 1) + x_mod],
+                            control[:, idx : idx + u_mod + 1],
                             data_params_in_dyn,
                             **self.plot_func[key][i].parameters,
                         )
+
+                        if self.plot_func[key][i].compute_derivative:
+                            # This is a special case since derivative is not properly integrated
+                            val = np.repeat(val, y_tp.shape[1])[np.newaxis, :]
+
+                        if val.shape != y_tp.shape:
+                            raise RuntimeError(
+                                f"Wrong dimensions for plot {key}. Got {val.shape}, but expected {y_tp.shape}"
+                            )
+                        y_tp[:, :] = val
                         all_y.append(y_tp)
 
                     for idx in range(len(self.plot_func[key][i].phase_mappings.map_idx)):
@@ -575,23 +654,57 @@ class PlotOcp:
                         for y in all_y:
                             y_tp.append(y[idx, :])
                         self.__append_to_ydata([y_tp])
+
+                elif self.plot_func[key][i].type == PlotType.POINT:
+                    for i_var in range(self.variable_sizes[i][key]):
+                        y = np.empty((len(self.plot_func[key][i].node_idx),))
+                        y.fill(np.nan)
+                        mod = 1 if self.plot_func[key][i].compute_derivative else 0
+                        for i_node, node_idx in enumerate(self.plot_func[key][i].node_idx):
+                            val = self.plot_func[key][i].function(
+                                node_idx,
+                                state[:, node_idx * step_size : (node_idx + 1) * step_size + mod : step_size],
+                                control[:, node_idx : node_idx + 1 + mod],
+                                data_params_in_dyn,
+                                **self.plot_func[key][i].parameters,
+                            )
+                            y[i_node] = val[i_var]
+                        self.ydata.append(y)
+
                 else:
                     y = np.empty((self.variable_sizes[i][key], len(self.t[i])))
                     y.fill(np.nan)
-                    try:
-                        y[:, :] = self.plot_func[key][i].function(
-                            state[:, ::step_size], control, data_params_in_dyn, **self.plot_func[key][i].parameters
-                        )
-                    except ValueError:
-                        val = (
-                            self.plot_func[key][i]
-                            .function(
-                                state[:, ::step_size], control, data_params_in_dyn, **self.plot_func[key][i].parameters
+                    if self.plot_func[key][i].compute_derivative:
+                        for i_node, node_idx in enumerate(self.plot_func[key][i].node_idx):
+                            val = self.plot_func[key][i].function(
+                                node_idx,
+                                state[:, node_idx * step_size : (node_idx + 1) * step_size + 1 : step_size],
+                                control[:, node_idx : node_idx + 1 + 1],
+                                data_params_in_dyn,
+                                **self.plot_func[key][i].parameters,
                             )
-                            .shape
+                            y[:, i_node] = val
+                    else:
+                        nodes = self.plot_func[key][i].node_idx
+                        if nodes and len(nodes) > 1 and len(nodes) == round(state.shape[1] / step_size):
+                            # Assume we are integrating but did not specify plot as such.
+                            # Therefore the arrival point is missing
+                            nodes += [nodes[-1] + 1]
+
+                        val = self.plot_func[key][i].function(
+                            nodes,
+                            state[:, ::step_size],
+                            control,
+                            data_params_in_dyn,
+                            **self.plot_func[key][i].parameters,
                         )
-                        raise ValueError(f"Wrong dimensions for plot {key}. Got " f"{val}" f", but expected {y.shape}")
+                        if val.shape != y.shape:
+                            raise RuntimeError(
+                                f"Wrong dimensions for plot {key}. Got {val.shape}, but expected {y.shape}"
+                            )
+                        y[:, :] = val
                     self.__append_to_ydata(y)
+
         self.__update_axes()
 
     def __update_xdata(self):
@@ -606,6 +719,9 @@ class PlotOcp:
                 for cmp, p in enumerate(plot[2]):
                     p.set_xdata(self.t_integrated[phase_idx][cmp])
                 ax = plot[2][-1].axes
+            elif plot[0] == PlotType.POINT:
+                plot[2].set_xdata(self.t[phase_idx][np.array(self.plot_func[plot[3]][phase_idx].node_idx)])
+                ax = plot[2].axes
             else:
                 plot[2].set_xdata(self.t[phase_idx])
                 ax = plot[2].axes
@@ -641,7 +757,6 @@ class PlotOcp:
         """
         Update the plotted data from ydata
         """
-
         assert len(self.plots) == len(self.ydata)
         for i, plot in enumerate(self.plots):
             y = self.ydata[i]
@@ -656,15 +771,15 @@ class PlotOcp:
             p.set_ydata((np.nan, np.nan))
 
         for key in self.axes:
-            if (not self.adapt_graph_size_to_bounds) or (self.axes[key][0].bounds is None):
+            if (not self.show_bounds) or (self.axes[key][0].bounds is None):
                 for i, ax in enumerate(self.axes[key][1]):
                     if not self.axes[key][0].ylim:
                         y_max = -np.inf
                         y_min = np.inf
                         for p in ax.get_children():
                             if isinstance(p, lines.Line2D):
-                                y_min = min(y_min, np.min(p.get_ydata()))
-                                y_max = max(y_max, np.max(p.get_ydata()))
+                                y_min = min(y_min, np.nanmin(p.get_ydata()))
+                                y_max = max(y_max, np.nanmax(p.get_ydata()))
                         y_range, data_range = self.__compute_ylim(y_min, y_max, 1.25)
                         ax.set_ylim(y_range)
                         ax.set_yticks(
@@ -765,7 +880,7 @@ class OnlineCallback(Callback):
         Send the current data to the plotter
     """
 
-    def __init__(self, ocp, opts: dict = None):
+    def __init__(self, ocp, opts: dict = None, show_options: dict = None):
         """
         Parameters
         ----------
@@ -773,6 +888,8 @@ class OnlineCallback(Callback):
             A reference to the ocp to show
         opts: dict
             Option to AnimateCallback method of CasADi
+        show_options: dict
+            The options to pass to PlotOcp
         """
         if opts is None:
             opts = {}
@@ -785,7 +902,7 @@ class OnlineCallback(Callback):
 
         self.queue = mp.Queue()
         self.plotter = self.ProcessPlotter(self.ocp)
-        self.plot_process = mp.Process(target=self.plotter, args=(self.queue,), daemon=True)
+        self.plot_process = mp.Process(target=self.plotter, args=(self.queue, show_options), daemon=True)
         self.plot_process.start()
 
     @staticmethod
@@ -911,16 +1028,20 @@ class OnlineCallback(Callback):
 
             self.ocp = ocp
 
-        def __call__(self, pipe: mp.Queue):
+        def __call__(self, pipe: mp.Queue, show_options: dict):
             """
             Parameters
             ----------
             pipe: mp.Queue
                 The multiprocessing queue to evaluate
+            show_options: dict
+                The option to pass to PlotOcp
             """
 
+            if show_options is None:
+                show_options = {}
             self.pipe = pipe
-            self.plot = PlotOcp(self.ocp)
+            self.plot = PlotOcp(self.ocp, **show_options)
             timer = self.plot.all_figures[0].canvas.new_timer(interval=10)
             timer.add_callback(self.callback)
             timer.start()
