@@ -7,7 +7,7 @@ from casadi import horzcat, vertcat, sum1, sum2, nlpsol, SX, MX, reshape
 from .solver_interface import SolverInterface
 from ..gui.plot import OnlineCallback
 from ..limits.path_conditions import Bounds
-from ..misc.enums import InterpolationType, ControlType
+from ..misc.enums import InterpolationType, ControlType, Node
 from ..optimization.solution import Solution
 
 
@@ -66,7 +66,7 @@ class IpoptInterface(SolverInterface):
         self.lam_g = None
         self.lam_x = None
 
-    def online_optim(self, ocp):
+    def online_optim(self, ocp, show_options: dict = None):
         """
         Declare the online callback to update the graphs while optimizing
 
@@ -74,10 +74,13 @@ class IpoptInterface(SolverInterface):
         ----------
         ocp: OptimalControlProgram
             A reference to the current OptimalControlProgram
+        show_options: dict
+            The options to pass to PlotOcp
         """
+
         if platform == "win32":
             raise RuntimeError("Online graphics are not available on Windows")
-        self.options_common["iteration_callback"] = OnlineCallback(ocp)
+        self.options_common["iteration_callback"] = OnlineCallback(ocp, show_options=show_options)
 
     def configure(self, solver_options: dict):
         """
@@ -257,8 +260,11 @@ class IpoptInterface(SolverInterface):
                     x_tp, u_tp = get_x_and_u_at_idx(penalty, idx)
                     x = horzcat(x, x_tp)
                     u = horzcat(u, u_tp)
-                if (penalty.derivative or penalty.explicit_derivative) and nlp.control_type == ControlType.CONSTANT:
+                if (
+                    penalty.derivative or penalty.explicit_derivative or penalty.node[0] == Node.ALL
+                ) and nlp.control_type == ControlType.CONSTANT:
                     u = horzcat(u, u[:, -1])
+
                 p = reshape(penalty.weighted_function(x, u, param, penalty.weight, target, penalty.dt), -1, 1)
 
             else:
