@@ -114,28 +114,8 @@ class RecedingHorizonOptimization(OptimalControlProgram):
             states.append(sol.states["all"][:, 0:1])
             controls.append(sol.controls["all"][:, 0:1])
 
-            # Update the initial frame bounds
-            if self.nlp[0].x_bounds.type != InterpolationType.CONSTANT_WITH_FIRST_AND_LAST_DIFFERENT:
-                if self.nlp[0].x_bounds.type == InterpolationType.CONSTANT:
-                    x_min = np.repeat(self.nlp[0].x_bounds.min[:, 0:1], 3, axis=1)
-                    x_max = np.repeat(self.nlp[0].x_bounds.max[:, 0:1], 3, axis=1)
-                    self.nlp[0].x_bounds = Bounds(x_min, x_max)
-                else:
-                    raise NotImplementedError(
-                        "The MHE is not implemented yet for x_bounds not being "
-                        "CONSTANT or CONSTANT_WITH_FIRST_AND_LAST_DIFFERENT"
-                    )
-                self.nlp[0].x_bounds.check_and_adjust_dimensions(self.nlp[0].states.shape, 3)
-            self.nlp[0].x_bounds[:, 0] = sol.states["all"][:, 1]
-
-            if self.nlp[0].x_init.type != InterpolationType.EACH_FRAME:
-                self.nlp[0].x_init = InitialGuess(
-                    np.ndarray(sol.states["all"].shape), interpolation=InterpolationType.EACH_FRAME
-                )
-                self.nlp[0].x_init.check_and_adjust_dimensions(self.nlp[0].states.shape, self.nlp[0].ns)
-            self.nlp[0].x_init.init[:, :] = np.concatenate(
-                (sol.states["all"][:, 1:], sol.states["all"][:, -1][:, np.newaxis]), axis=1
-            )
+            # Update the initial frame bounds and initial guess
+            self._advance_windows(sol)
 
             t += 1
         real_time = time() - real_time
@@ -156,6 +136,30 @@ class RecedingHorizonOptimization(OptimalControlProgram):
         sol.real_time_to_optimize = real_time
         return sol
 
+    def _advance_windows(self, sol: Solution, steps: int = 0):
+        # Update the initial frame bounds
+        if self.nlp[0].x_bounds.type != InterpolationType.CONSTANT_WITH_FIRST_AND_LAST_DIFFERENT:
+            if self.nlp[0].x_bounds.type == InterpolationType.CONSTANT:
+                x_min = np.repeat(self.nlp[0].x_bounds.min[:, 0:1], 3, axis=1)
+                x_max = np.repeat(self.nlp[0].x_bounds.max[:, 0:1], 3, axis=1)
+                self.nlp[0].x_bounds = Bounds(x_min, x_max)
+            else:
+                raise NotImplementedError(
+                    "The MHE is not implemented yet for x_bounds not being "
+                    "CONSTANT or CONSTANT_WITH_FIRST_AND_LAST_DIFFERENT"
+                )
+            self.nlp[0].x_bounds.check_and_adjust_dimensions(self.nlp[0].states.shape, 3)
+        self.nlp[0].x_bounds[:, 0] = sol.states["all"][:, 1]
+
+        if self.nlp[0].x_init.type != InterpolationType.EACH_FRAME:
+            self.nlp[0].x_init = InitialGuess(
+                np.ndarray(sol.states["all"].shape), interpolation=InterpolationType.EACH_FRAME
+            )
+            self.nlp[0].x_init.check_and_adjust_dimensions(self.nlp[0].states.shape, self.nlp[0].ns)
+        self.nlp[0].x_init.init[:, :] = np.concatenate(
+            (sol.states["all"][:, 1:], sol.states["all"][:, -1][:, np.newaxis]), axis=1
+        )
+
     def _define_time(
         self,
         phase_time: Union[int, float, list, tuple],
